import { world, MinecraftItemTypes, ItemStack, Items, MinecraftEnchantmentTypes, Enchantment, BlockLocation, EntityQueryOptions } from 'mojang-minecraft'
import { ActionFormData, ModalFormData } from 'mojang-minecraft-ui'

function enchantListToArray(enchantList) {
    let arr = []

    for (const enchant in MinecraftEnchantmentTypes) {
        let level = enchantList.hasEnchantment(MinecraftEnchantmentTypes[enchant])
        if (level > 0) {
            arr.push({
                id: enchant,
                level: level
            })
        }
    }

    return arr
}

function enchantArrayToString(enchantArr) { // enchantArr being one generated by enchantListToArray
    let strArr = []

    for (const enchant of enchantArr) {
        strArr.push(`${enchant.id} ${enchant.level}`)
    }

    return strArr.join(', ')
}

function addItemToStorage(storageEntity, item) {
    let tag = storageEntity.getTags()[0]
    let json = JSON.parse(tag)

    let itemKey = JSON.stringify([item.id, item.data, item.nameTag, item.getLore(), enchantListToArray(item.getComponent('enchantments').enchantments)]) //TODO: add enchantments list here // use these constants to tell the uniqueness of items

    if (json[itemKey]) {
        json[itemKey].amount += item.amount
    }
    else {
        json[itemKey] = {
            amount: item.amount
        }
    }

    storageEntity.removeTag(tag)
    storageEntity.addTag(JSON.stringify(json))
}

function setStorageItemAmount(storageEntity, itemKey, newAmount) {
    let tag = storageEntity.getTags()[0]
    let json = JSON.parse(tag)

    if (newAmount > 0) {
        json[itemKey].amount = newAmount
    }
    else {
        delete json[itemKey]
    }

    storageEntity.removeTag(tag)
    storageEntity.addTag(JSON.stringify(json))
}

function spawnItemStack(player, parsedKey, amountRequested) {
    for (let i = 0; i < amountRequested; i++) { // since we dont know the max stack size of the item, we'll have to do it one by one
        let item = new ItemStack(Items.get(parsedKey[0]), 1, parsedKey[1])
        item.nameTag = parsedKey[2]
        item.setLore(parsedKey[3])

        let enchants = item.getComponent('enchantments')
        let enchantments = enchants.enchantments
        for (const enchant of parsedKey[4]) {
            enchantments.addEnchantment(new Enchantment(MinecraftEnchantmentTypes[enchant.id], enchant.level))
        }
        enchants.enchantments = enchantments // no clue why this works

        let inventory = player.getComponent('inventory').container
        if (inventory.emptySlotsCount > 0) {
            inventory.addItem(item)
        }
        else {
            player.dimension.spawnItem(item, player.location)
        }
    }
}

function showStorageUI(player, storageEntity) {
    let items = Object.entries(JSON.parse(storageEntity.getTags()[0]))

    let menu = new ActionFormData()
        .title('Storage Terminal')

    if (items.length > 0) {

        let itemNames = []

        for (const [key, item] of items) {
            let keyArr = JSON.parse(key)

            let name = `${keyArr[0].split(':')[1].replaceAll('_', ' ')}${(keyArr[1] > 0) ? ` (${keyArr[1]})` : ''}${(keyArr[2]) ? ` "${keyArr[2]}"` : ''}${(keyArr[4].length > 0) ? `\n§8${enchantArrayToString(keyArr[4])}` : ''}${(keyArr[3].length > 0) ? `\n§5${keyArr[3].join(', ')}` : ''}`
            itemNames.push(name)

            menu.button(`${item.amount}x ${name}`)
        }

        menu.show(player).then(response => {
            if (!response.isCanceled) {
                let item = items[response.selection][1]
                let key = items[response.selection][0]
                let parsedKey = JSON.parse(key)

                new ModalFormData()
                    .title('Storage Terminal')
                    .slider(`Requesting:\n${itemNames[response.selection]}\n§r--\nHow many of this item to request?`, 1, item.amount, 1, Math.min(Math.round(item.amount / 2), 64))
                    .show(player).then(modalResponse => {
                        if (!modalResponse.isCanceled) {
                            let amountRequested = modalResponse.formValues[0]
                            spawnItemStack(player, parsedKey, amountRequested)
                            setStorageItemAmount(storageEntity, key, item.amount - amountRequested)
                        }

                        showStorageUI(player, storageEntity)
                    })
            }
        })
    }
    else {
        menu.body('No items to display. Sneak+Punch while holding an item to add it')
        menu.button('Ok')
        menu.show(player)
    }
}

world.events.entityHit.subscribe(e => {

    // interacting with terminal
    if (e.entity.id == 'minecraft:player' && e.hitEntity && e.hitEntity.id == 'astoragenet:storage_terminal') {
        if (e.entity.isSneaking) { // if sneaking put held item into storage
            let inventory = e.entity.getComponent('inventory').container
            let heldItem = inventory.getItem(e.entity.selectedSlot)
            if (heldItem) {
                if (heldItem.id == 'astoragenet:used_storage_drive') {
                    if (e.hitEntity.getTags()[0] == '{}') { // test if there is any data in the terminal
                        if (heldItem.getLore().length == 1) {
                            e.hitEntity.removeTag('{}')
                            e.hitEntity.addTag(heldItem.getLore()[0])
                            inventory.setItem(e.entity.selectedSlot, new ItemStack(MinecraftItemTypes.air, 0))
                        }
                        else {
                            e.entity.runCommand('tellraw @s { "rawtext": [ { "text": "§cCannot input drive, drive has none or invalid data!" } ] }')
                        }
                    }
                    else {
                        e.entity.runCommand('tellraw @s { "rawtext": [ { "text": "§cCan only input drive into empty terminals!" } ] }')
                    }
                }
                else {
                    inventory.setItem(e.entity.selectedSlot, new ItemStack(MinecraftItemTypes.air, 0))
                    addItemToStorage(e.hitEntity, heldItem)
                }
            }
        }
        else {
            showStorageUI(e.entity, e.hitEntity)
        }
    }
})

world.events.dataDrivenEntityTriggerEvent.subscribe(e => {

    // picking up terminal
    if (e.id == 'astoragenet:pickup' && e.entity.id == 'astoragenet:storage_terminal') {
        if (e.entity.getTags()[0] != '{}') {
            let item = new ItemStack(Items.get('astoragenet:used_storage_drive'))
            item.setLore([e.entity.getTags()[0]])
            e.entity.dimension.spawnItem(item, e.entity.location)
        }
        e.entity.dimension.spawnItem(new ItemStack(Items.get('astoragenet:storage_terminal')), e.entity.location)
    }

    // import bus
    else if (e.id == 'astoragenet:tick' && e.entity.id == 'astoragenet:terminal_import_bus') {
        let dimension = e.entity.dimension
        let chest = dimension.getBlock(new BlockLocation(e.entity.location.x, e.entity.location.y + 1, e.entity.location.z))
        if (chest.getComponent('inventory')) { // check if block has an inventory (is a chest-like object)
            let inventory = chest.getComponent('inventory').container

            let query = new EntityQueryOptions()
            query.location = e.entity.location
            query.closest = 1
            query.maxDistance = 5
            query.families = ['astoragenet:storage_terminal']
            let terminal = [...dimension.getEntities(query)][0]

            if (terminal) {
                for (let i = 0; i < inventory.size - 1; i++) {
                    let item = inventory.getItem(i)
                    if (item) {
                        inventory.setItem(i, new ItemStack(MinecraftItemTypes.air, 0))
                        addItemToStorage(terminal, item)
                    }
                }
            }
        }
    }
})